# 23 "asllib/Lexer.mll"
 

open Tokens

open Error

module type CONFIG = sig
    (** Allow variables starting with a double underscore (__) *)
    val allow_double_underscore : bool
    val allow_unknown : bool
    val allow_single_arrows : bool
    val allow_function_like_statements : bool
end

let reserved_keywords = []

let is_reserved_keyword: string -> bool =
  let tbl: (string, unit) Hashtbl.t = Hashtbl.create (List.length reserved_keywords) in
  let () = List.iter (fun s -> Hashtbl.add tbl s ()) reserved_keywords in
  fun s -> Hashtbl.mem tbl s

(* Get a Tokens.token from the string name of the token.
   A useful utility when working with cmly files.

   Note that this set only contains simple enumerative types, not compound types

   This function's exhaustiveness is guaranteed by the test in tests/Lexer. *)
let token_of_string =
 let s t = Some t in
 function
 | "ACCESSOR"           -> s ACCESSOR
 | "AND"                -> s AND
 | "ARRAY"              -> s ARRAY
 | "ARROW"              -> s ARROW
 | "AS"                 -> s AS
 | "ASSERT"             -> s ASSERT
 | "BAND"               -> s BAND
 | "BEGIN"              -> s BEGIN
 | "BEQ"                -> s BEQ
 | "BIT"                -> s BIT
 | "BITS"               -> s BITS
 | "BNOT"               -> s BNOT
 | "BOOLEAN"            -> s BOOLEAN
 | "BOR"                -> s BOR
 | "CASE"               -> s CASE
 | "CATCH"              -> s CATCH
 | "COLLECTION"         -> s COLLECTION
 | "COLON"              -> s COLON
 | "COLON_COLON"        -> s COLON_COLON
 | "COMMA"              -> s COMMA
 | "CONFIG"             -> s CONFIG
 | "CONSTANT"           -> s CONSTANT
 | "DEBUG"              -> s DEBUG
 | "DIV"                -> s DIV
 | "DIVRM"              -> s DIVRM
 | "DO"                 -> s DO
 | "DOT"                -> s DOT
 | "DOWNTO"             -> s DOWNTO
 | "ELSE"               -> s ELSE
 | "ELSIF"              -> s ELSIF
 | "END"                -> s END
 | "ENUMERATION"        -> s ENUMERATION
 | "EOF"                -> s EOF
 | "XOR"                -> s XOR
 | "EQ"                 -> s EQ
 | "EQ_OP"              -> s EQ_OP
 | "EXCEPTION"          -> s EXCEPTION
 | "FOR"                -> s FOR
 | "FUNC"               -> s FUNC
 | "GEQ"                -> s GEQ
 | "GETTER"             -> s GETTER
 | "GT"                 -> s GT
 | "IF"                 -> s IF
 | "IMPLEMENTATION"     -> s IMPLEMENTATION
 | "IMPDEF"             -> s IMPDEF
 | "IMPL"               -> s IMPL
 | "IN"                 -> s IN
 | "INTEGER"            -> s INTEGER
 | "LBRACE"             -> s LBRACE
 | "LBRACKET"           -> s LBRACKET
 | "LEQ"                -> s LEQ
 | "LET"                -> s LET
 | "LOOPLIMIT"          -> s LOOPLIMIT
 | "LPAR"               -> s LPAR
 | "LT"                 -> s LT
 | "MINUS"              -> s MINUS
 | "MOD"                -> s MOD
 | "MUL"                -> s MUL
 | "NEQ"                -> s NEQ
 | "NORETURN"           -> s NORETURN
 | "NOT"                -> s NOT
 | "OF"                 -> s OF
 | "OR"                 -> s OR
 | "OTHERWISE"          -> s OTHERWISE
 | "PASS"               -> s PASS
 | "PLUS"               -> s PLUS
 | "PLUS_COLON"         -> s PLUS_COLON
 | "POW"                -> s POW
 | "PRAGMA"             -> s PRAGMA
 | "PRINTLN"            -> s PRINTLN
 | "PRINT"              -> s PRINT
 | "PURE"               -> s PURE
 | "RBRACE"             -> s RBRACE
 | "RBRACKET"           -> s RBRACKET
 | "RDIV"               -> s RDIV
 | "READONLY"           -> s READONLY
 | "REAL"               -> s REAL
 | "RECORD"             -> s RECORD
 | "RECURSELIMIT"       -> s RECURSELIMIT
 | "REPEAT"             -> s REPEAT
 | "RETURN"             -> s RETURN
 | "RPAR"               -> s RPAR
 | "STAR_COLON"         -> s STAR_COLON
 | "SEMI_COLON"         -> s SEMI_COLON
 | "SETTER"             -> s SETTER
 | "SHL"                -> s SHL
 | "SHR"                -> s SHR
 | "SLICING"            -> s SLICING
 | "STRING"             -> s STRING
 | "SUBTYPES"           -> s SUBTYPES
 | "THEN"               -> s THEN
 | "THROW"              -> s THROW
 | "TO"                 -> s TO
 | "TRY"                -> s TRY
 | "TYPE"               -> s TYPE
 | "ARBITRARY"          -> s ARBITRARY
 | "UNREACHABLE"        -> s UNREACHABLE
 | "UNTIL"              -> s UNTIL
 | "VAR"                -> s VAR
 | "WHEN"               -> s WHEN
 | "WHERE"              -> s WHERE
 | "WHILE"              -> s WHILE
 | "WITH"               -> s WITH
 | "LLBRACKET"          -> s LLBRACKET
 | "RRBRACKET"          -> s RRBRACKET
 | _ -> None

(** Convert a lexical token to the symbol it lexes *)
let token_to_symbol = function
  | BNOT               -> "!"
  | COMMA              -> ","
  | LT                 -> "<"
  | SHR                -> ">>"
  | BAND               -> "&&"
  | IMPL               -> "==>"
  | SHL                -> "<<"
  | RBRACKET           -> "]"
  | RRBRACKET          -> "]]"
  | RPAR               -> ")"
  | SLICING            -> ".."
  | EQ                 -> "="
  | LBRACE             -> "{"
  | NEQ                -> "!="
  | MINUS              -> "-"
  | BEQ                -> "<=>"
  | LBRACKET           -> "["
  | LLBRACKET          -> "[["
  | LPAR               -> "("
  | DOT                -> "."
  | LEQ                -> "<="
  | POW                -> "^"
  | MUL                -> "*"
  | RDIV               -> "/"
  | EQ_OP              -> "=="
  | BOR                -> "||"
  | PLUS               -> "+"
  | COLON              -> ":"
  | ARROW              -> "=>"
  | RBRACE             -> "}"
  | COLON_COLON        -> "::"
  | GT                 -> ">"
  | PLUS_COLON         -> "+:"
  | STAR_COLON         -> "*:"
  | SEMI_COLON         -> ";"
  | GEQ                -> ">="
  (* Keywords *)
  | ACCESSOR           -> "accessor"
  | AND                -> "AND"
  | ARRAY              -> "array"
  | AS                 -> "as"
  | ASSERT             -> "assert"
  | BEGIN              -> "begin"
  | BIT                -> "bit"
  | BITS               -> "bits"
  | BOOLEAN            -> "boolean"
  | CASE               -> "case"
  | CATCH              -> "catch"
  | COLLECTION         -> "collection"
  | CONFIG             -> "config"
  | CONSTANT           -> "constant"
  | DIV                -> "DIV"
  | DIVRM              -> "DIVRM"
  | DO                 -> "do"
  | DOWNTO             -> "downto"
  | ELSE               -> "else"
  | ELSIF              -> "elsif"
  | END                -> "end"
  | ENUMERATION        -> "enumeration"
  | XOR                -> "XOR"
  | EXCEPTION          -> "exception"
  | FOR                -> "for"
  | FUNC               -> "func"
  | GETTER             -> "getter"
  | IF                 -> "if"
  | IMPLEMENTATION     -> "implementation"
  | IMPDEF             -> "impdef"
  | IN                 -> "IN"
  | INTEGER            -> "integer"
  | LET                -> "let"
  | LOOPLIMIT          -> "looplimit"
  | MOD                -> "MOD"
  | NORETURN           -> "noreturn"
  | NOT                -> "NOT"
  | OF                 -> "of"
  | OR                 -> "OR"
  | OTHERWISE          -> "otherwise"
  | PASS               -> "pass"
  | PRAGMA             -> "pragma"
  | PRINTLN            -> "println"
  | PRINT              -> "print"
  | PURE               -> "pure"
  | READONLY           -> "readonly"
  | REAL               -> "real"
  | RECORD             -> "record"
  | RECURSELIMIT       -> "recurselimit"
  | REPEAT             -> "repeat"
  | RETURN             -> "return"
  | SETTER             -> "setter"
  | STRING             -> "string"
  | SUBTYPES           -> "subtypes"
  | THEN               -> "then"
  | THROW              -> "throw"
  | TO                 -> "to"
  | TRY                -> "try"
  | TYPE               -> "type"
  | ARBITRARY          -> "ARBITRARY"
  | UNREACHABLE        -> "unreachable"
  | UNTIL              -> "until"
  | VAR                -> "var"
  | WHEN               -> "when"
  | WHERE              -> "where"
  | WHILE              -> "while"
  | WITH               -> "with"
  | BOOL_LIT _
  | INT_LIT _
  | REAL_LIT _
  | STRING_LIT _
  | BITVECTOR_LIT _
  | MASK_LIT _
  | IDENTIFIER _
  | EOF
  | DEBUG -> assert false (* Complex tokens *)


module Make (Config : CONFIG) = struct

exception LexerError

let new_line lexbuf = Lexing.new_line lexbuf; lexbuf
let bitvector_lit lxm = BITVECTOR_LIT (Bitvector.of_string lxm)
let mask_lit lxm = MASK_LIT (Bitvector.mask_of_string lxm)
let reserved_err s = Error.fatal_unknown_pos @@ (Error.ReservedIdentifier s)

let fatal lexbuf desc =
  AST.
    {
      desc;
      version = V1;
      pos_start = Lexing.lexeme_start_p lexbuf;
      pos_end = Lexing.lexeme_end_p lexbuf;
    }
  |> Error.fatal

let tr_name s = match s with
| "accessor"      -> ACCESSOR
| "AND"           -> AND
| "array"         -> ARRAY
| "as"            -> AS
| "assert"        -> ASSERT
| "begin"         -> BEGIN
| "bit"           -> BIT
| "bits"          -> BITS
| "boolean"       -> BOOLEAN
| "case"          -> CASE
| "catch"         -> CATCH
| "collection"    -> COLLECTION
| "config"        -> CONFIG
| "constant"      -> CONSTANT
| "__debug__"
| "__DEBUG__"     -> DEBUG
| "DIV"           -> DIV
| "DIVRM"         -> DIVRM
| "do"            -> DO
| "downto"        -> DOWNTO
| "else"          -> ELSE
| "elsif"         -> ELSIF
| "end"           -> END
| "enumeration"   -> ENUMERATION
| "XOR"           -> XOR
| "exception"     -> EXCEPTION
| "FALSE"         -> BOOL_LIT false
| "for"           -> FOR
| "func"          -> FUNC
| "getter"        -> GETTER
| "if"            -> IF
| "impdef"        -> IMPDEF
| "implementation" -> IMPLEMENTATION
| "IN"            -> IN
| "integer"       -> INTEGER
| "let"           -> LET
| "looplimit"     -> LOOPLIMIT
| "MOD"           -> MOD
| "NOT"           -> NOT
| "noreturn"      -> NORETURN
| "of"            -> OF
| "OR"            -> OR
| "otherwise"     -> OTHERWISE
| "pass"          -> PASS
| "pragma"        -> PRAGMA
| "println"       -> PRINTLN
| "print"         -> PRINT
| "pure"          -> PURE
| "readonly"      -> READONLY
| "real"          -> REAL
| "record"        -> RECORD
| "recurselimit"  -> RECURSELIMIT
| "repeat"        -> REPEAT
| "return"        -> RETURN
| "setter"        -> SETTER
| "string"        -> STRING
| "subtypes"      -> SUBTYPES
| "then"          -> THEN
| "throw"         -> THROW
| "to"            -> TO
| "try"           -> TRY
| "TRUE"          -> BOOL_LIT true
| "type"          -> TYPE
| "UNKNOWN"       ->
    if Config.allow_unknown then ARBITRARY
    else fatal_unknown_pos (Error.ObsoleteSyntax s)
| "ARBITRARY"     -> ARBITRARY
| "Unreachable"   ->
    if Config.allow_function_like_statements then UNREACHABLE
    else fatal_unknown_pos (Error.ObsoleteSyntax s)
| "unreachable"   -> UNREACHABLE
| "until"         -> UNTIL
| "var"           -> VAR
| "when"          -> WHEN
| "where"         -> WHERE
| "while"         -> WHILE
| "with"          -> WITH
(* Reserved identifiers *)
| x when is_reserved_keyword x -> reserved_err x
| x when not Config.allow_double_underscore
         && ASTUtils.string_starts_with ~prefix:"__" x -> reserved_err x
(* End of reserved identifiers *)
| x               -> IDENTIFIER x

# 361 "asllib/Lexer.ml"
let __ocaml_lex_tables = {
  Lexing.lex_base =
   "\000\000\251\255\252\255\253\255\254\255\255\255\001\000\002\000\
    \253\255\254\255\255\255\003\000\004\000\253\255\000\000\255\255\
    \086\000\204\255\196\000\000\000\208\255\214\255\000\000\001\000\
    \025\000\002\000\222\255\225\255\000\000\232\255\002\000\235\255\
    \004\000\000\000\004\000\000\000\002\000\005\000\244\255\006\000\
    \031\001\248\255\022\000\039\001\009\000\007\000\255\255\252\255\
    \031\000\166\000\055\001\201\255\119\001\200\255\203\255\215\001\
    \202\255\005\000\050\002\247\255\246\255\057\000\231\255\007\000\
    \018\000\238\255\229\255\228\255\207\255\242\255\241\255\019\000\
    \240\255\215\255\020\000\239\255\236\255\234\255\226\255\209\255\
    \218\255\210\255\213\255\212\255\000\000\001\000\001\000\001\000\
    \002\000\005\000\010\000\005\000\206\255";
  Lexing.lex_backtrk =
   "\255\255\255\255\255\255\255\255\255\255\255\255\004\000\003\000\
    \255\255\255\255\255\255\004\000\003\000\255\255\001\000\255\255\
    \056\000\255\255\050\000\255\255\255\255\255\255\039\000\038\000\
    \255\255\034\000\255\255\255\255\028\000\255\255\031\000\255\255\
    \018\000\022\000\025\000\255\255\044\000\012\000\255\255\010\000\
    \255\255\255\255\004\000\004\000\035\000\001\000\255\255\255\255\
    \002\000\255\255\255\255\255\255\005\000\255\255\255\255\006\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\032\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\036\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255";
  Lexing.lex_default =
   "\001\000\000\000\000\000\000\000\000\000\000\000\007\000\007\000\
    \000\000\000\000\000\000\012\000\012\000\000\000\255\255\000\000\
    \255\255\000\000\255\255\255\255\000\000\000\000\255\255\255\255\
    \255\255\255\255\000\000\000\000\255\255\000\000\255\255\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\255\255\
    \255\255\000\000\255\255\255\255\255\255\255\255\000\000\000\000\
    \048\000\054\000\051\000\000\000\255\255\000\000\000\000\255\255\
    \000\000\255\255\255\255\000\000\000\000\255\255\000\000\255\255\
    \255\255\000\000\000\000\000\000\000\000\000\000\000\000\255\255\
    \000\000\000\000\255\255\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\000\000";
  Lexing.lex_trans =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\008\000\255\255\013\000\255\255\000\000\
    \045\000\000\000\000\000\000\000\045\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\003\000\010\000\255\255\061\000\070\000\045\000\
    \000\000\255\255\000\000\000\000\082\000\014\000\255\255\015\000\
    \077\000\071\000\064\000\047\000\000\000\061\000\061\000\000\000\
    \048\000\000\000\083\000\081\000\079\000\074\000\073\000\068\000\
    \069\000\065\000\063\000\062\000\049\000\067\000\042\000\042\000\
    \042\000\042\000\042\000\042\000\042\000\042\000\042\000\042\000\
    \066\000\072\000\075\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\061\000\000\000\078\000\002\000\009\000\255\255\045\000\
    \046\000\076\000\058\000\045\000\000\000\000\000\000\000\000\000\
    \000\000\061\000\061\000\089\000\084\000\088\000\005\000\085\000\
    \086\000\087\000\090\000\091\000\004\000\042\000\045\000\039\000\
    \041\000\092\000\000\000\000\000\035\000\040\000\027\000\031\000\
    \025\000\023\000\038\000\034\000\030\000\044\000\043\000\042\000\
    \042\000\042\000\042\000\042\000\042\000\042\000\042\000\042\000\
    \022\000\020\000\037\000\033\000\036\000\080\000\019\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\028\000\000\000\032\000\026\000\018\000\000\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\029\000\024\000\021\000\255\255\000\000\055\000\055\000\
    \055\000\055\000\055\000\055\000\055\000\055\000\055\000\055\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\000\000\000\000\
    \255\255\255\255\255\255\255\255\255\255\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\255\255\
    \000\000\000\000\000\000\018\000\000\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\040\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\059\000\057\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\040\000\
    \040\000\000\000\000\000\000\000\000\000\049\000\017\000\042\000\
    \042\000\042\000\042\000\042\000\042\000\042\000\042\000\042\000\
    \042\000\000\000\000\000\000\000\000\000\000\000\000\000\052\000\
    \052\000\052\000\052\000\052\000\052\000\052\000\052\000\052\000\
    \052\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \052\000\052\000\052\000\052\000\052\000\052\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\042\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\058\000\
    \052\000\052\000\052\000\052\000\052\000\052\000\000\000\050\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\255\255\052\000\
    \052\000\052\000\052\000\052\000\052\000\052\000\052\000\052\000\
    \052\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \052\000\052\000\052\000\052\000\052\000\052\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\000\000\000\000\000\000\000\000\052\000\000\000\
    \052\000\052\000\052\000\052\000\052\000\052\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\055\000\
    \055\000\055\000\055\000\055\000\055\000\055\000\055\000\055\000\
    \055\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \056\000\056\000\056\000\056\000\056\000\056\000\056\000\056\000\
    \056\000\056\000\056\000\056\000\056\000\056\000\056\000\056\000\
    \056\000\056\000\056\000\056\000\056\000\056\000\056\000\056\000\
    \056\000\056\000\000\000\000\000\000\000\000\000\055\000\255\255\
    \056\000\056\000\056\000\056\000\056\000\056\000\056\000\056\000\
    \056\000\056\000\056\000\056\000\056\000\056\000\056\000\056\000\
    \056\000\056\000\056\000\056\000\056\000\056\000\056\000\056\000\
    \056\000\056\000\058\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\060\000\057\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\058\000\058\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\058\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000";
  Lexing.lex_check =
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\006\000\007\000\011\000\012\000\255\255\
    \045\000\255\255\255\255\255\255\045\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\006\000\007\000\057\000\035\000\045\000\
    \255\255\048\000\255\255\255\255\023\000\011\000\012\000\014\000\
    \030\000\034\000\037\000\044\000\255\255\057\000\057\000\255\255\
    \044\000\255\255\022\000\023\000\025\000\033\000\033\000\036\000\
    \036\000\037\000\037\000\039\000\042\000\063\000\042\000\042\000\
    \042\000\042\000\042\000\042\000\042\000\042\000\042\000\042\000\
    \064\000\071\000\074\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\061\000\255\255\028\000\000\000\006\000\007\000\016\000\
    \016\000\032\000\061\000\016\000\255\255\255\255\255\255\255\255\
    \255\255\061\000\061\000\088\000\019\000\087\000\000\000\084\000\
    \085\000\086\000\089\000\090\000\000\000\042\000\016\000\016\000\
    \016\000\091\000\255\255\255\255\016\000\016\000\016\000\016\000\
    \016\000\016\000\016\000\016\000\016\000\016\000\016\000\016\000\
    \016\000\016\000\016\000\016\000\016\000\016\000\016\000\016\000\
    \016\000\016\000\016\000\016\000\016\000\024\000\016\000\016\000\
    \016\000\016\000\016\000\016\000\016\000\016\000\016\000\016\000\
    \016\000\016\000\016\000\016\000\016\000\016\000\016\000\016\000\
    \016\000\016\000\016\000\016\000\016\000\016\000\016\000\016\000\
    \016\000\016\000\255\255\016\000\016\000\016\000\255\255\016\000\
    \016\000\016\000\016\000\016\000\016\000\016\000\016\000\016\000\
    \016\000\016\000\016\000\016\000\016\000\016\000\016\000\016\000\
    \016\000\016\000\016\000\016\000\016\000\016\000\016\000\016\000\
    \016\000\016\000\016\000\016\000\049\000\255\255\049\000\049\000\
    \049\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\255\255\255\255\
    \000\000\006\000\007\000\011\000\012\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\048\000\
    \255\255\255\255\255\255\018\000\255\255\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\040\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\040\000\040\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\040\000\
    \040\000\255\255\255\255\255\255\255\255\043\000\016\000\043\000\
    \043\000\043\000\043\000\043\000\043\000\043\000\043\000\043\000\
    \043\000\255\255\255\255\255\255\255\255\255\255\255\255\050\000\
    \050\000\050\000\050\000\050\000\050\000\050\000\050\000\050\000\
    \050\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \050\000\050\000\050\000\050\000\050\000\050\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\043\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\040\000\
    \050\000\050\000\050\000\050\000\050\000\050\000\255\255\043\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\049\000\052\000\
    \052\000\052\000\052\000\052\000\052\000\052\000\052\000\052\000\
    \052\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \052\000\052\000\052\000\052\000\052\000\052\000\052\000\052\000\
    \052\000\052\000\052\000\052\000\052\000\052\000\052\000\052\000\
    \052\000\052\000\052\000\052\000\052\000\052\000\052\000\052\000\
    \052\000\052\000\255\255\255\255\255\255\255\255\052\000\255\255\
    \052\000\052\000\052\000\052\000\052\000\052\000\052\000\052\000\
    \052\000\052\000\052\000\052\000\052\000\052\000\052\000\052\000\
    \052\000\052\000\052\000\052\000\052\000\052\000\052\000\052\000\
    \052\000\052\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\055\000\
    \055\000\055\000\055\000\055\000\055\000\055\000\055\000\055\000\
    \055\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \055\000\055\000\055\000\055\000\055\000\055\000\055\000\055\000\
    \055\000\055\000\055\000\055\000\055\000\055\000\055\000\055\000\
    \055\000\055\000\055\000\055\000\055\000\055\000\055\000\055\000\
    \055\000\055\000\255\255\255\255\255\255\255\255\055\000\050\000\
    \055\000\055\000\055\000\055\000\055\000\055\000\055\000\055\000\
    \055\000\055\000\055\000\055\000\055\000\055\000\055\000\055\000\
    \055\000\055\000\055\000\055\000\055\000\055\000\055\000\055\000\
    \055\000\055\000\058\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\058\000\058\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\058\000\058\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\058\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255";
  Lexing.lex_base_code =
   "";
  Lexing.lex_backtrk_code =
   "";
  Lexing.lex_default_code =
   "";
  Lexing.lex_trans_code =
   "";
  Lexing.lex_check_code =
   "";
  Lexing.lex_code =
   "";
}

let rec escaped_string_chars acc lexbuf =
   __ocaml_lex_escaped_string_chars_rec acc lexbuf 0
and __ocaml_lex_escaped_string_chars_rec acc lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 424 "asllib/Lexer.mll"
         ( Buffer.add_char acc '\n'; string_lit acc lexbuf )
# 631 "asllib/Lexer.ml"

  | 1 ->
# 425 "asllib/Lexer.mll"
         ( Buffer.add_char acc '\t'; string_lit acc lexbuf )
# 636 "asllib/Lexer.ml"

  | 2 ->
# 426 "asllib/Lexer.mll"
         ( Buffer.add_char acc '"'; string_lit acc lexbuf )
# 641 "asllib/Lexer.ml"

  | 3 ->
# 427 "asllib/Lexer.mll"
         ( Buffer.add_char acc '\\'; string_lit acc lexbuf )
# 646 "asllib/Lexer.ml"

  | 4 ->
# 428 "asllib/Lexer.mll"
                         ( raise LexerError )
# 651 "asllib/Lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_escaped_string_chars_rec acc lexbuf __ocaml_lex_state

and string_lit acc lexbuf =
   __ocaml_lex_string_lit_rec acc lexbuf 6
and __ocaml_lex_string_lit_rec acc lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 431 "asllib/Lexer.mll"
          ( STRING_LIT (Buffer.contents acc) )
# 663 "asllib/Lexer.ml"

  | 1 ->
# 432 "asllib/Lexer.mll"
          ( escaped_string_chars acc lexbuf )
# 668 "asllib/Lexer.ml"

  | 2 ->
# 433 "asllib/Lexer.mll"
          ( Buffer.add_char acc '\n'; new_line lexbuf |> string_lit acc )
# 673 "asllib/Lexer.ml"

  | 3 ->
let
# 434 "asllib/Lexer.mll"
                          lxm
# 679 "asllib/Lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 434 "asllib/Lexer.mll"
                              ( Buffer.add_string acc lxm; string_lit acc lexbuf )
# 683 "asllib/Lexer.ml"

  | 4 ->
# 435 "asllib/Lexer.mll"
          ( raise LexerError )
# 688 "asllib/Lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_string_lit_rec acc lexbuf __ocaml_lex_state

and c_comments lexbuf =
   __ocaml_lex_c_comments_rec lexbuf 11
and __ocaml_lex_c_comments_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 443 "asllib/Lexer.mll"
                  ( token      lexbuf )
# 700 "asllib/Lexer.ml"

  | 1 ->
# 444 "asllib/Lexer.mll"
                  ( c_comments lexbuf )
# 705 "asllib/Lexer.ml"

  | 2 ->
# 445 "asllib/Lexer.mll"
                  ( new_line lexbuf |> c_comments )
# 710 "asllib/Lexer.ml"

  | 3 ->
# 446 "asllib/Lexer.mll"
                  ( c_comments lexbuf )
# 715 "asllib/Lexer.ml"

  | 4 ->
# 447 "asllib/Lexer.mll"
                  ( raise LexerError  )
# 720 "asllib/Lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_c_comments_rec lexbuf __ocaml_lex_state

and token lexbuf =
   __ocaml_lex_token_rec lexbuf 16
and __ocaml_lex_token_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 455 "asllib/Lexer.mll"
                               ( new_line lexbuf |> token         )
# 732 "asllib/Lexer.ml"

  | 1 ->
# 456 "asllib/Lexer.mll"
                               ( token lexbuf                     )
# 737 "asllib/Lexer.ml"

  | 2 ->
# 457 "asllib/Lexer.mll"
                               ( token lexbuf                     )
# 742 "asllib/Lexer.ml"

  | 3 ->
# 458 "asllib/Lexer.mll"
                               ( c_comments lexbuf                )
# 747 "asllib/Lexer.ml"

  | 4 ->
let
# 459 "asllib/Lexer.mll"
                 lxm
# 753 "asllib/Lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 459 "asllib/Lexer.mll"
                               ( INT_LIT(Z.of_string lxm)         )
# 757 "asllib/Lexer.ml"

  | 5 ->
let
# 460 "asllib/Lexer.mll"
                 lxm
# 763 "asllib/Lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 460 "asllib/Lexer.mll"
                               ( INT_LIT(Z.of_string lxm)         )
# 767 "asllib/Lexer.ml"

  | 6 ->
let
# 461 "asllib/Lexer.mll"
                  lxm
# 773 "asllib/Lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 461 "asllib/Lexer.mll"
                               ( REAL_LIT(Q.of_string lxm)        )
# 777 "asllib/Lexer.ml"

  | 7 ->
# 462 "asllib/Lexer.mll"
                               ( string_lit (Buffer.create 16) lexbuf )
# 782 "asllib/Lexer.ml"

  | 8 ->
let
# 463 "asllib/Lexer.mll"
                    lxm
# 788 "asllib/Lexer.ml"
= Lexing.sub_lexeme lexbuf (lexbuf.Lexing.lex_start_pos + 1) (lexbuf.Lexing.lex_curr_pos + -1) in
# 463 "asllib/Lexer.mll"
                               ( bitvector_lit lxm                )
# 792 "asllib/Lexer.ml"

  | 9 ->
let
# 464 "asllib/Lexer.mll"
                    lxm
# 798 "asllib/Lexer.ml"
= Lexing.sub_lexeme lexbuf (lexbuf.Lexing.lex_start_pos + 1) (lexbuf.Lexing.lex_curr_pos + -1) in
# 464 "asllib/Lexer.mll"
                               ( mask_lit lxm                     )
# 802 "asllib/Lexer.ml"

  | 10 ->
# 465 "asllib/Lexer.mll"
                               ( BNOT                             )
# 807 "asllib/Lexer.ml"

  | 11 ->
# 466 "asllib/Lexer.mll"
                               ( COMMA                            )
# 812 "asllib/Lexer.ml"

  | 12 ->
# 467 "asllib/Lexer.mll"
                               ( LT                               )
# 817 "asllib/Lexer.ml"

  | 13 ->
# 468 "asllib/Lexer.mll"
                               ( SHR                              )
# 822 "asllib/Lexer.ml"

  | 14 ->
# 469 "asllib/Lexer.mll"
                               ( BAND                             )
# 827 "asllib/Lexer.ml"

  | 15 ->
# 470 "asllib/Lexer.mll"
                               ( if Config.allow_single_arrows then IMPL else fatal lexbuf (ObsoleteSyntax "implication with -->") )
# 832 "asllib/Lexer.ml"

  | 16 ->
# 471 "asllib/Lexer.mll"
                               ( IMPL                             )
# 837 "asllib/Lexer.ml"

  | 17 ->
# 472 "asllib/Lexer.mll"
                               ( SHL                              )
# 842 "asllib/Lexer.ml"

  | 18 ->
# 473 "asllib/Lexer.mll"
                               ( RBRACKET                         )
# 847 "asllib/Lexer.ml"

  | 19 ->
# 474 "asllib/Lexer.mll"
                               ( RRBRACKET                        )
# 852 "asllib/Lexer.ml"

  | 20 ->
# 475 "asllib/Lexer.mll"
                               ( RPAR                             )
# 857 "asllib/Lexer.ml"

  | 21 ->
# 476 "asllib/Lexer.mll"
                               ( SLICING                          )
# 862 "asllib/Lexer.ml"

  | 22 ->
# 477 "asllib/Lexer.mll"
                               ( EQ                               )
# 867 "asllib/Lexer.ml"

  | 23 ->
# 478 "asllib/Lexer.mll"
                               ( LBRACE                           )
# 872 "asllib/Lexer.ml"

  | 24 ->
# 479 "asllib/Lexer.mll"
                               ( NEQ                              )
# 877 "asllib/Lexer.ml"

  | 25 ->
# 480 "asllib/Lexer.mll"
                               ( MINUS                            )
# 882 "asllib/Lexer.ml"

  | 26 ->
# 481 "asllib/Lexer.mll"
                               ( if Config.allow_single_arrows then BEQ else fatal lexbuf (ObsoleteSyntax "equivalence with <->") )
# 887 "asllib/Lexer.ml"

  | 27 ->
# 482 "asllib/Lexer.mll"
                               ( BEQ                              )
# 892 "asllib/Lexer.ml"

  | 28 ->
# 483 "asllib/Lexer.mll"
                               ( LBRACKET                         )
# 897 "asllib/Lexer.ml"

  | 29 ->
# 484 "asllib/Lexer.mll"
                               ( LLBRACKET                        )
# 902 "asllib/Lexer.ml"

  | 30 ->
# 485 "asllib/Lexer.mll"
                               ( LPAR                             )
# 907 "asllib/Lexer.ml"

  | 31 ->
# 486 "asllib/Lexer.mll"
                               ( DOT                              )
# 912 "asllib/Lexer.ml"

  | 32 ->
# 487 "asllib/Lexer.mll"
                               ( LEQ                              )
# 917 "asllib/Lexer.ml"

  | 33 ->
# 488 "asllib/Lexer.mll"
                               ( POW                              )
# 922 "asllib/Lexer.ml"

  | 34 ->
# 489 "asllib/Lexer.mll"
                               ( MUL                              )
# 927 "asllib/Lexer.ml"

  | 35 ->
# 490 "asllib/Lexer.mll"
                               ( RDIV                             )
# 932 "asllib/Lexer.ml"

  | 36 ->
# 491 "asllib/Lexer.mll"
                               ( EQ_OP                            )
# 937 "asllib/Lexer.ml"

  | 37 ->
# 492 "asllib/Lexer.mll"
                               ( BOR                              )
# 942 "asllib/Lexer.ml"

  | 38 ->
# 493 "asllib/Lexer.mll"
                               ( PLUS                             )
# 947 "asllib/Lexer.ml"

  | 39 ->
# 494 "asllib/Lexer.mll"
                               ( COLON                            )
# 952 "asllib/Lexer.ml"

  | 40 ->
# 495 "asllib/Lexer.mll"
                               ( ARROW                            )
# 957 "asllib/Lexer.ml"

  | 41 ->
# 496 "asllib/Lexer.mll"
                               ( RBRACE                           )
# 962 "asllib/Lexer.ml"

  | 42 ->
# 497 "asllib/Lexer.mll"
                               ( fatal lexbuf (ObsoleteSyntax "string concatenation with ++") )
# 967 "asllib/Lexer.ml"

  | 43 ->
# 498 "asllib/Lexer.mll"
                               ( COLON_COLON                      )
# 972 "asllib/Lexer.ml"

  | 44 ->
# 499 "asllib/Lexer.mll"
                               ( GT                               )
# 977 "asllib/Lexer.ml"

  | 45 ->
# 500 "asllib/Lexer.mll"
                               ( PLUS_COLON                       )
# 982 "asllib/Lexer.ml"

  | 46 ->
# 501 "asllib/Lexer.mll"
                               ( STAR_COLON                       )
# 987 "asllib/Lexer.ml"

  | 47 ->
# 502 "asllib/Lexer.mll"
                               ( SEMI_COLON                       )
# 992 "asllib/Lexer.ml"

  | 48 ->
# 503 "asllib/Lexer.mll"
                               ( GEQ                              )
# 997 "asllib/Lexer.ml"

  | 49 ->
# 504 "asllib/Lexer.mll"
                               ( fatal lexbuf (ObsoleteSyntax "Loop limits with @looplimit") )
# 1002 "asllib/Lexer.ml"

  | 50 ->
let
# 505 "asllib/Lexer.mll"
                    lxm
# 1008 "asllib/Lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 505 "asllib/Lexer.mll"
                               ( tr_name lxm                      )
# 1012 "asllib/Lexer.ml"

  | 51 ->
# 506 "asllib/Lexer.mll"
                               ( EOF                              )
# 1017 "asllib/Lexer.ml"

  | 52 ->
# 507 "asllib/Lexer.mll"
                               ( raise LexerError                 )
# 1022 "asllib/Lexer.ml"

  | 53 ->
# 508 "asllib/Lexer.mll"
                               ( raise LexerError                 )
# 1027 "asllib/Lexer.ml"

  | 54 ->
# 509 "asllib/Lexer.mll"
                               ( raise LexerError                 )
# 1032 "asllib/Lexer.ml"

  | 55 ->
# 510 "asllib/Lexer.mll"
                               ( raise LexerError                 )
# 1037 "asllib/Lexer.ml"

  | 56 ->
# 511 "asllib/Lexer.mll"
                               ( raise LexerError                 )
# 1042 "asllib/Lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_token_rec lexbuf __ocaml_lex_state

;;

# 512 "asllib/Lexer.mll"
 
end

# 1053 "asllib/Lexer.ml"
